@@ -1,7 +1,7 @@
 /* Buffer management for tar.
 
-   Copyright 1988, 1992, 1993, 1994, 1996, 1997, 1999, 2000, 2001 Free
-   Software Foundation, Inc.
+   Copyright (C) 1988, 1992, 1993, 1994, 1996, 1997, 1999, 2000, 2001,
+   2003, 2004 Free Software Foundation, Inc.
 
    Written by John Gilmore, on 1985-08-25.
 
@@ -23,14 +23,6 @@
 
 #include <signal.h>
 
-#if MSDOS
-# include <process.h>
-#endif
-
-#if XENIX
-# include <sys/inode.h>
-#endif
-
 #include <fnmatch.h>
 #include <human.h>
 #include <quotearg.h>
@@ -38,9 +30,6 @@
 #include "common.h"
 #include "rmt.h"
 
-#define	PREAD 0			/* read file descriptor from pipe() */
-#define	PWRITE 1		/* write file descriptor from pipe() */
-
 /* Number of retries before giving up on read.  */
 #define	READ_ERROR_MAX 10
 
@@ -51,6 +40,7 @@
 
 static tarlong prev_written;	/* bytes written on previous volumes */
 static tarlong bytes_written;	/* bytes written on this volume */
+static void *record_buffer;	/* allocated memory */
 
 /* FIXME: The following variables should ideally be static to this
    module.  However, this cannot be done yet.  The cleanup continues!  */
@@ -62,24 +52,13 @@ enum access_mode access_mode;	/* how do we handle the archive */
 off_t records_read;		/* number of records read from this archive */
 off_t records_written;		/* likewise, for records written */
 
-static struct stat archive_stat; /* stat block for archive file */
-
 static off_t record_start_block; /* block ordinal at record_start */
 
-/* Where we write list messages (not errors, not interactions) to.  Stdout
-   unless we're writing a pipe, in which case stderr.  */
+/* Where we write list messages (not errors, not interactions) to.  */
 FILE *stdlis;
 
-static void backspace_output PARAMS ((void));
-static int new_volume PARAMS ((enum access_mode));
-static void archive_write_error PARAMS ((ssize_t)) __attribute__ ((noreturn));
-static void archive_read_error PARAMS ((void));
-
-#if !MSDOS
-/* Obnoxious test to see if dimwit is trying to dump the archive.  */
-dev_t ar_dev;
-ino_t ar_ino;
-#endif
+static void backspace_output (void);
+static bool new_volume (enum access_mode);
 
 /* PID of child program, if compress_option or remote archive access.  */
 static pid_t child_pid;
@@ -93,12 +72,13 @@ static int hit_eof;
 /* Checkpointing counter */
 static int checkpoint;
 
-/* We're reading, but we just read the last block and its time to update.  */
-/* As least EXTERN like this one as possible.  FIXME!  */
-extern int time_to_start_writing;
+/* We're reading, but we just read the last block and it's time to update.
+   Declared in update.c
 
-int file_to_switch_to = -1;	/* if remote update, close archive, and use
-				   this descriptor to write to */
+   As least EXTERN like this one as possible. (?? --gray)
+   FIXME: Either eliminate it or move it to common.h.
+*/
+extern bool time_to_start_writing;
 
 static int volno = 1;		/* which volume of a multi-volume tape we're
 				   on */
@@ -129,6 +109,42 @@ static off_t real_s_sizeleft;
 
 /* Functions.  */
 
+void
+clear_read_error_count (void)
+{
+  read_error_count = 0;
+}
+
+
+/* Time-related functions */
+
+double duration;
+
+void
+set_start_time ()
+{
+#if HAVE_CLOCK_GETTIME
+  if (clock_gettime (CLOCK_REALTIME, &start_timespec) != 0)
+#endif
+    start_time = time (0);
+}
+
+void
+compute_duration ()
+{
+#if HAVE_CLOCK_GETTIME
+  struct timespec now;
+  if (clock_gettime (CLOCK_REALTIME, &now) == 0)
+    duration += ((now.tv_sec - start_timespec.tv_sec)
+		 + (now.tv_nsec - start_timespec.tv_nsec) / 1e9);
+  else
+#endif
+    duration += time (NULL) - start_time;
+  set_start_time ();
+}
+
+
+
 void
 print_total_written (void)
 {
@@ -136,24 +152,16 @@ print_total_written (void)
   char bytes[sizeof (tarlong) * CHAR_BIT];
   char abbr[LONGEST_HUMAN_READABLE + 1];
   char rate[LONGEST_HUMAN_READABLE + 1];
-  double seconds;
-
-#if HAVE_CLOCK_GETTIME
-  struct timespec now;
-  if (clock_gettime (CLOCK_REALTIME, &now) == 0)
-    seconds = ((now.tv_sec - start_timespec.tv_sec)
-	       + (now.tv_nsec - start_timespec.tv_nsec) / 1e9);
-  else
-#endif
-    seconds = time (0) - start_time;
+  
+  int human_opts = human_autoscale | human_base_1024 | human_SI | human_B;
 
   sprintf (bytes, TARLONG_FORMAT, written);
 
   /* Amanda 2.4.1p1 looks for "Total bytes written: [0-9][0-9]*".  */
-  fprintf (stderr, _("Total bytes written: %s (%sB, %sB/s)\n"), bytes,
-	   human_readable ((uintmax_t) written, abbr, 1, -1024),
-	   (0 < seconds && written / seconds < (uintmax_t) -1
-	    ? human_readable ((uintmax_t) (written / seconds), rate, 1, -1024)
+  fprintf (stderr, _("Total bytes written: %s (%s, %s/s)\n"), bytes,
+	   human_readable (written, abbr, human_opts, 1, 1),
+	   (0 < duration && written / duration < (uintmax_t) -1
+	    ? human_readable (written / duration, rate, human_opts, 1, 1)
 	    : "?"));
 }
 
@@ -197,8 +205,7 @@ find_next_block (void)
   return current_block;
 }
 
-/* Indicate that we have used all blocks up thru BLOCK.
-   FIXME: should the arg have an off-by-1?  */
+/* Indicate that we have used all blocks up thru BLOCK. */
 void
 set_next_block_after (union block *block)
 {
@@ -216,7 +223,7 @@ set_next_block_after (union block *block)
 /* Return the number of bytes comprising the space between POINTER
    through the end of the current buffer of blocks.  This space is
    available for filling with data, or taking data from.  POINTER is
-   usually (but not always) the result previous find_next_block call.  */
+   usually (but not always) the result of previous find_next_block call.  */
 size_t
 available_space_after (union block *pointer)
 {
@@ -224,421 +231,31 @@ available_space_after (union block *pointer)
 }
 
 /* Close file having descriptor FD, and abort if close unsuccessful.  */
-static void
+void
 xclose (int fd)
 {
   if (close (fd) != 0)
     close_error (_("(pipe)"));
 }
 
-/* Duplicate file descriptor FROM into becoming INTO.
-   INTO is closed first and has to be the next available slot.  */
-static void
-xdup2 (int from, int into)
-{
-  if (from != into)
-    {
-      int status = close (into);
-
-      if (status != 0 && errno != EBADF)
-	{
-	  int e = errno;
-	  FATAL_ERROR ((0, e, _("Cannot close")));
-	}
-      status = dup (from);
-      if (status != into)
-	{
-	  if (status < 0)
-	    {
-	      int e = errno;
-	      FATAL_ERROR ((0, e, _("Cannot dup")));
-	    }
-	  abort ();
-	}
-      xclose (from);
-    }
-}
-
-#if MSDOS
-
-/* Set ARCHIVE for writing, then compressing an archive.  */
-static void
-child_open_for_compress (void)
-{
-  FATAL_ERROR ((0, 0, _("Cannot use compressed or remote archives")));
-}
-
-/* Set ARCHIVE for uncompressing, then reading an archive.  */
-static void
-child_open_for_uncompress (void)
-{
-  FATAL_ERROR ((0, 0, _("Cannot use compressed or remote archives")));
-}
-
-#else /* not MSDOS */
-
-/* Return nonzero if NAME is the name of a regular file, or if the file
-   does not exist (so it would be created as a regular file).  */
-static int
-is_regular_file (const char *name)
-{
-  struct stat stbuf;
-
-  if (stat (name, &stbuf) == 0)
-    return S_ISREG (stbuf.st_mode);
-  else
-    return errno == ENOENT;
-}
-
-static ssize_t
-write_archive_buffer (void)
-{
-  ssize_t status;
-  ssize_t written = 0;
-
-  while (0 <= (status = rmtwrite (archive, record_start->buffer + written,
-				  record_size - written)))
-    {
-      written += status;
-      if (written == record_size
-	  || _isrmt (archive)
-	  || ! (S_ISFIFO (archive_stat.st_mode)
-		|| S_ISSOCK (archive_stat.st_mode)))
-	break;
-    }
-
-  return written ? written : status;
-}
-
-/* Set ARCHIVE for writing, then compressing an archive.  */
-static void
-child_open_for_compress (void)
-{
-  int parent_pipe[2];
-  int child_pipe[2];
-  pid_t grandchild_pid;
-  int wait_status;
-
-  xpipe (parent_pipe);
-  child_pid = xfork ();
-
-  if (child_pid > 0)
-    {
-      /* The parent tar is still here!  Just clean up.  */
-
-      archive = parent_pipe[PWRITE];
-      xclose (parent_pipe[PREAD]);
-      return;
-    }
-
-  /* The new born child tar is here!  */
-
-  program_name = _("tar (child)");
-
-  xdup2 (parent_pipe[PREAD], STDIN_FILENO);
-  xclose (parent_pipe[PWRITE]);
-
-  /* Check if we need a grandchild tar.  This happens only if either:
-     a) we are writing stdout: to force reblocking;
-     b) the file is to be accessed by rmt: compressor doesn't know how;
-     c) the file is not a plain file.  */
-
-  if (strcmp (archive_name_array[0], "-") != 0
-      && !_remdev (archive_name_array[0])
-      && is_regular_file (archive_name_array[0]))
-    {
-      if (backup_option)
-	maybe_backup_file (archive_name_array[0], 1);
-
-      /* We don't need a grandchild tar.  Open the archive and launch the
-	 compressor.  */
-
-      archive = creat (archive_name_array[0], MODE_RW);
-      if (archive < 0)
-	{
-	  int saved_errno = errno;
-
-	  if (backup_option)
-	    undo_last_backup ();
-	  errno = saved_errno;
-	  open_fatal (archive_name_array[0]);
-	}
-      xdup2 (archive, STDOUT_FILENO);
-      execlp (use_compress_program_option, use_compress_program_option,
-	      (char *) 0);
-      exec_fatal (use_compress_program_option);
-    }
-
-  /* We do need a grandchild tar.  */
-
-  xpipe (child_pipe);
-  grandchild_pid = xfork ();
-
-  if (grandchild_pid == 0)
-    {
-      /* The newborn grandchild tar is here!  Launch the compressor.  */
-
-      program_name = _("tar (grandchild)");
-
-      xdup2 (child_pipe[PWRITE], STDOUT_FILENO);
-      xclose (child_pipe[PREAD]);
-      execlp (use_compress_program_option, use_compress_program_option,
-	      (char *) 0);
-      exec_fatal (use_compress_program_option);
-    }
-
-  /* The child tar is still here!  */
-
-  /* Prepare for reblocking the data from the compressor into the archive.  */
-
-  xdup2 (child_pipe[PREAD], STDIN_FILENO);
-  xclose (child_pipe[PWRITE]);
-
-  if (strcmp (archive_name_array[0], "-") == 0)
-    archive = STDOUT_FILENO;
-  else
-    {
-      archive = rmtcreat (archive_name_array[0], MODE_RW, rsh_command_option);
-      if (archive < 0)
-	open_fatal (archive_name_array[0]);
-    }
-
-  /* Let's read out of the stdin pipe and write an archive.  */
-
-  while (1)
-    {
-      ssize_t status = 0;
-      char *cursor;
-      size_t length;
-
-      /* Assemble a record.  */
-
-      for (length = 0, cursor = record_start->buffer;
-	   length < record_size;
-	   length += status, cursor += status)
-	{
-	  size_t size = record_size - length;
-
-	  if (size < BLOCKSIZE)
-	    size = BLOCKSIZE;
-	  status = safe_read (STDIN_FILENO, cursor, size);
-	  if (status <= 0)
-	    break;
-	}
-
-      if (status < 0)
-	read_fatal (use_compress_program_option);
-
-      /* Copy the record.  */
-
-      if (status == 0)
-	{
-	  /* We hit the end of the file.  Write last record at
-	     full length, as the only role of the grandchild is
-	     doing proper reblocking.  */
-
-	  if (length > 0)
-	    {
-	      memset (record_start->buffer + length, 0, record_size - length);
-	      status = write_archive_buffer ();
-	      if (status != record_size)
-		archive_write_error (status);
-	    }
-
-	  /* There is nothing else to read, break out.  */
-	  break;
-	}
-
-      status = write_archive_buffer ();
-      if (status != record_size)
-	archive_write_error (status);
-    }
-
-#if 0
-  close_archive ();
-#endif
-
-  /* Propagate any failure of the grandchild back to the parent.  */
-
-  while (waitpid (grandchild_pid, &wait_status, 0) == -1)
-    if (errno != EINTR)
-      {
-	waitpid_error (use_compress_program_option);
-	break;
-      }
-
-  if (WIFSIGNALED (wait_status))
-    {
-      kill (child_pid, WTERMSIG (wait_status));
-      exit_status = TAREXIT_FAILURE;
-    }
-  else if (WEXITSTATUS (wait_status) != 0)
-    exit_status = WEXITSTATUS (wait_status);
-
-  exit (exit_status);
-}
-
-/* Set ARCHIVE for uncompressing, then reading an archive.  */
-static void
-child_open_for_uncompress (void)
-{
-  int parent_pipe[2];
-  int child_pipe[2];
-  pid_t grandchild_pid;
-  int wait_status;
-
-  xpipe (parent_pipe);
-  child_pid = xfork ();
-
-  if (child_pid > 0)
-    {
-      /* The parent tar is still here!  Just clean up.  */
-
-      read_full_records_option = 1;
-      archive = parent_pipe[PREAD];
-      xclose (parent_pipe[PWRITE]);
-      return;
-    }
-
-  /* The newborn child tar is here!  */
-
-  program_name = _("tar (child)");
-
-  xdup2 (parent_pipe[PWRITE], STDOUT_FILENO);
-  xclose (parent_pipe[PREAD]);
-
-  /* Check if we need a grandchild tar.  This happens only if either:
-     a) we're reading stdin: to force unblocking;
-     b) the file is to be accessed by rmt: compressor doesn't know how;
-     c) the file is not a plain file.  */
-
-  if (strcmp (archive_name_array[0], "-") != 0
-      && !_remdev (archive_name_array[0])
-      && is_regular_file (archive_name_array[0]))
-    {
-      /* We don't need a grandchild tar.  Open the archive and lauch the
-	 uncompressor.  */
-
-      archive = open (archive_name_array[0], O_RDONLY | O_BINARY, MODE_RW);
-      if (archive < 0)
-	open_fatal (archive_name_array[0]);
-      xdup2 (archive, STDIN_FILENO);
-      execlp (use_compress_program_option, use_compress_program_option,
-	      "-d", (char *) 0);
-      exec_fatal (use_compress_program_option);
-    }
-
-  /* We do need a grandchild tar.  */
-
-  xpipe (child_pipe);
-  grandchild_pid = xfork ();
-
-  if (grandchild_pid == 0)
-    {
-      /* The newborn grandchild tar is here!  Launch the uncompressor.  */
-
-      program_name = _("tar (grandchild)");
-
-      xdup2 (child_pipe[PREAD], STDIN_FILENO);
-      xclose (child_pipe[PWRITE]);
-      execlp (use_compress_program_option, use_compress_program_option,
-	      "-d", (char *) 0);
-      exec_fatal (use_compress_program_option);
-    }
-
-  /* The child tar is still here!  */
-
-  /* Prepare for unblocking the data from the archive into the
-     uncompressor.  */
-
-  xdup2 (child_pipe[PWRITE], STDOUT_FILENO);
-  xclose (child_pipe[PREAD]);
-
-  if (strcmp (archive_name_array[0], "-") == 0)
-    archive = STDIN_FILENO;
-  else
-    archive = rmtopen (archive_name_array[0], O_RDONLY | O_BINARY,
-		       MODE_RW, rsh_command_option);
-  if (archive < 0)
-    open_fatal (archive_name_array[0]);
-
-  /* Let's read the archive and pipe it into stdout.  */
-
-  while (1)
-    {
-      char *cursor;
-      size_t maximum;
-      size_t count;
-      ssize_t status;
-
-      read_error_count = 0;
-
-    error_loop:
-      status = rmtread (archive, record_start->buffer, record_size);
-      if (status < 0)
-	{
-	  archive_read_error ();
-	  goto error_loop;
-	}
-      if (status == 0)
-	break;
-      cursor = record_start->buffer;
-      maximum = status;
-      while (maximum)
-	{
-	  count = maximum < BLOCKSIZE ? maximum : BLOCKSIZE;
-	  if (full_write (STDOUT_FILENO, cursor, count) != count)
-	    write_error (use_compress_program_option);
-	  cursor += count;
-	  maximum -= count;
-	}
-    }
-
-  xclose (STDOUT_FILENO);
-#if 0
-  close_archive ();
-#endif
-
-  /* Propagate any failure of the grandchild back to the parent.  */
-
-  while (waitpid (grandchild_pid, &wait_status, 0) == -1)
-    if (errno != EINTR)
-      {
-	waitpid_error (use_compress_program_option);
-	break;
-      }
-
-  if (WIFSIGNALED (wait_status))
-    {
-      kill (child_pid, WTERMSIG (wait_status));
-      exit_status = TAREXIT_FAILURE;
-    }
-  else if (WEXITSTATUS (wait_status) != 0)
-    exit_status = WEXITSTATUS (wait_status);
-
-  exit (exit_status);
-}
-
-#endif /* not MSDOS */
-
 /* Check the LABEL block against the volume label, seen as a globbing
    pattern.  Return true if the pattern matches.  In case of failure,
    retry matching a volume sequence number before giving up in
    multi-volume mode.  */
-static int
+static bool
 check_label_pattern (union block *label)
 {
   char *string;
-  int result;
+  bool result;
 
   if (! memchr (label->header.name, '\0', sizeof label->header.name))
-    return 0;
+    return false;
 
   if (fnmatch (volume_label_option, label->header.name, 0) == 0)
-    return 1;
+    return true;
 
   if (!multi_volume_option)
-    return 0;
+    return false;
 
   string = xmalloc (strlen (volume_label_option)
 		    + sizeof VOLUME_LABEL_APPEND + 1);
@@ -656,7 +273,14 @@ open_archive (enum access_mode wanted_access)
 {
   int backed_up_flag = 0;
 
-  stdlis = to_stdout_option ? stderr : stdout;
+  if (index_file_name)
+    {
+      stdlis = fopen (index_file_name, "w");
+      if (! stdlis)
+	open_error (index_file_name);
+    }
+  else
+    stdlis = to_stdout_option ? stderr : stdout;
 
   if (record_size == 0)
     FATAL_ERROR ((0, 0, _("Invalid value for record_size")));
@@ -664,24 +288,16 @@ open_archive (enum access_mode wanted_access)
   if (archive_names == 0)
     FATAL_ERROR ((0, 0, _("No archive name given")));
 
-  current_file_name = 0;
-  current_link_name = 0;
+  tar_stat_destroy (&current_stat_info);
   save_name = 0;
   real_s_name = 0;
 
+  record_start =
+    page_aligned_alloc (&record_buffer,
+			(record_size
+			 + (multi_volume_option ? 2 * BLOCKSIZE : 0)));
   if (multi_volume_option)
-    {
-      if (verify_option)
-	FATAL_ERROR ((0, 0, _("Cannot verify multi-volume archives")));
-      record_start = valloc (record_size + (2 * BLOCKSIZE));
-      if (record_start)
-	record_start += 2;
-    }
-  else
-    record_start = valloc (record_size);
-  if (!record_start)
-    FATAL_ERROR ((0, 0, _("Cannot allocate memory for blocking factor %d"),
-		  blocking_factor));
+    record_start += 2;
 
   current_block = record_start;
   record_end = record_start + blocking_factor;
@@ -690,23 +306,18 @@ open_archive (enum access_mode wanted_access)
 
   if (use_compress_program_option)
     {
-      if (multi_volume_option)
-	FATAL_ERROR ((0, 0, _("Cannot use multi-volume compressed archives")));
-      if (verify_option)
-	FATAL_ERROR ((0, 0, _("Cannot verify compressed archives")));
-
       switch (wanted_access)
 	{
 	case ACCESS_READ:
-	  child_open_for_uncompress ();
+	  child_pid = sys_child_open_for_uncompress ();
 	  break;
 
 	case ACCESS_WRITE:
-	  child_open_for_compress ();
+	  child_pid = sys_child_open_for_compress ();
 	  break;
 
 	case ACCESS_UPDATE:
-	  FATAL_ERROR ((0, 0, _("Cannot update compressed archives")));
+	  abort (); /* Should not happen */
 	  break;
 	}
 
@@ -716,7 +327,7 @@ open_archive (enum access_mode wanted_access)
     }
   else if (strcmp (archive_name_array[0], "-") == 0)
     {
-      read_full_records_option = 1; /* could be a pipe, be safe */
+      read_full_records_option = true; /* could be a pipe, be safe */
       if (verify_option)
 	FATAL_ERROR ((0, 0, _("Cannot verify stdin/stdout archive")));
 
@@ -734,7 +345,7 @@ open_archive (enum access_mode wanted_access)
 	case ACCESS_UPDATE:
 	  archive = STDIN_FILENO;
 	  stdlis = stderr;
-	  write_archive_to_stdout = 1;
+	  write_archive_to_stdout = true;
 	  break;
 	}
     }
@@ -766,7 +377,7 @@ open_archive (enum access_mode wanted_access)
       }
 
   if (archive < 0
-      || (! _isrmt (archive) && fstat (archive, &archive_stat) < 0))
+      || (! _isrmt (archive) && !sys_get_archive_stat ()))
     {
       int saved_errno = errno;
 
@@ -776,35 +387,9 @@ open_archive (enum access_mode wanted_access)
       open_fatal (archive_name_array[0]);
     }
 
-#if !MSDOS
-
-  /* Detect if outputting to "/dev/null".  */
-  {
-    static char const dev_null[] = "/dev/null";
-    struct stat dev_null_stat;
-
-    dev_null_output =
-      (strcmp (archive_name_array[0], dev_null) == 0
-       || (! _isrmt (archive)
-	   && S_ISCHR (archive_stat.st_mode)
-	   && stat (dev_null, &dev_null_stat) == 0
-	   && archive_stat.st_dev == dev_null_stat.st_dev
-	   && archive_stat.st_ino == dev_null_stat.st_ino));
-  }
-
-  if (!_isrmt (archive) && S_ISREG (archive_stat.st_mode))
-    {
-      ar_dev = archive_stat.st_dev;
-      ar_ino = archive_stat.st_ino;
-    }
-  else
-    ar_dev = 0;
-
-#endif /* not MSDOS */
-
-#if MSDOS
-  setmode (archive, O_BINARY);
-#endif
+  sys_detect_dev_null_output ();
+  sys_save_archive_dev_ino ();
+  SET_BINARY_MODE (archive);
 
   switch (wanted_access)
     {
@@ -840,14 +425,14 @@ open_archive (enum access_mode wanted_access)
 	  else
 	    strcpy (record_start->header.name, volume_label_option);
 
-	  assign_string (&current_file_name, record_start->header.name);
+	  assign_string (&current_stat_info.file_name,
+			 record_start->header.name);
+	  current_stat_info.had_trailing_slash =
+	    strip_trailing_slashes (current_stat_info.file_name);
 
 	  record_start->header.typeflag = GNUTYPE_VOLHDR;
 	  TIME_TO_CHARS (start_time, record_start->header.mtime);
-	  finish_header (record_start);
-#if 0
-	  current_block++;
-#endif
+	  finish_header (&current_stat_info, record_start, -1);
 	}
       break;
     }
@@ -871,7 +456,7 @@ flush_write (void)
   else if (dev_null_output)
     status = record_size;
   else
-    status = write_archive_buffer ();
+    status = sys_write_archive_buffer ();
   if (status != record_size && !multi_volume_option)
     archive_write_error (status);
 
@@ -885,23 +470,18 @@ flush_write (void)
     {
       if (multi_volume_option)
 	{
-	  char *cursor;
-
-	  if (!save_name)
+	  if (save_name)
+	    {
+	      assign_string (&real_s_name, safer_name_suffix (save_name, false));
+	      real_s_totsize = save_totsize;
+	      real_s_sizeleft = save_sizeleft;
+	    }
+	  else
 	    {
 	      assign_string (&real_s_name, 0);
 	      real_s_totsize = 0;
 	      real_s_sizeleft = 0;
-	      return;
 	    }
-
-	  cursor = save_name + FILESYSTEM_PREFIX_LEN (save_name);
-	  while (ISSLASH (*cursor))
-	    cursor++;
-
-	  assign_string (&real_s_name, cursor);
-	  real_s_totsize = save_totsize;
-	  real_s_sizeleft = save_sizeleft;
 	}
       return;
     }
@@ -941,7 +521,7 @@ flush_write (void)
 	       volume_label_option, volno);
       TIME_TO_CHARS (start_time, record_start->header.mtime);
       record_start->header.typeflag = GNUTYPE_VOLHDR;
-      finish_header (record_start);
+      finish_header (&current_stat_info, record_start, -1);
     }
 
   if (real_s_name)
@@ -951,27 +531,32 @@ flush_write (void)
       if (volume_label_option)
 	record_start++;
 
+      if (strlen (real_s_name) > NAME_FIELD_SIZE)
+	FATAL_ERROR ((0, 0,
+		      _("%s: file name too long to be stored in a GNU multivolume header"),
+		      quotearg_colon (real_s_name)));
+      
       memset (record_start, 0, BLOCKSIZE);
 
       /* FIXME: Michael P Urban writes: [a long name file] is being written
 	 when a new volume rolls around [...]  Looks like the wrong value is
 	 being preserved in real_s_name, though.  */
 
-      strcpy (record_start->header.name, real_s_name);
+      strncpy (record_start->header.name, real_s_name, NAME_FIELD_SIZE);
       record_start->header.typeflag = GNUTYPE_MULTIVOL;
       OFF_TO_CHARS (real_s_sizeleft, record_start->header.size);
       OFF_TO_CHARS (real_s_totsize - real_s_sizeleft,
 		    record_start->oldgnu_header.offset);
       tmp = verbose_option;
       verbose_option = 0;
-      finish_header (record_start);
+      finish_header (&current_stat_info, record_start, -1);
       verbose_option = tmp;
 
       if (volume_label_option)
 	record_start--;
     }
 
-  status = write_archive_buffer ();
+  status = sys_write_archive_buffer ();
   if (status != record_size)
     archive_write_error (status);
 
@@ -991,12 +576,7 @@ flush_write (void)
 	assign_string (&real_s_name, 0);
       else
 	{
-	  char *cursor = save_name + FILESYSTEM_PREFIX_LEN (save_name);
-
-	  while (ISSLASH (*cursor))
-	    cursor++;
-
-	  assign_string (&real_s_name, cursor);
+	  assign_string (&real_s_name, safer_name_suffix (save_name, false));
 	  real_s_sizeleft = save_sizeleft;
 	  real_s_totsize = save_totsize;
 	}
@@ -1007,7 +587,7 @@ flush_write (void)
 /* Handle write errors on the archive.  Write errors are always fatal.
    Hitting the end of a volume does not cause a write error unless the
    write was the first record of the volume.  */
-static void
+void
 archive_write_error (ssize_t status)
 {
   /* It might be useful to know how much was written before the error
@@ -1024,7 +604,7 @@ archive_write_error (ssize_t status)
 
 /* Handle read errors on the archive.  If the read should be retried,
    return to the caller.  */
-static void
+void
 archive_read_error (void)
 {
   read_error (*archive_name_cursor);
@@ -1040,13 +620,72 @@ archive_read_error (void)
   return;
 }
 
+static void
+short_read (size_t status)
+{
+  size_t left;			/* bytes left */
+  char *more;			/* pointer to next byte to read */
+
+  more = record_start->buffer + status;
+  left = record_size - status;
+
+  while (left % BLOCKSIZE != 0
+	 || (left && status && read_full_records_option))
+    {
+      if (status)
+	while ((status = rmtread (archive, more, left)) == SAFE_READ_ERROR)
+	  archive_read_error ();
+
+      if (status == 0)
+	{
+	  char buf[UINTMAX_STRSIZE_BOUND];
+
+	  WARN((0, 0, _("Read %s bytes from %s"),
+		STRINGIFY_BIGINT (record_size - left, buf),
+		*archive_name_cursor));
+	  break;
+	}
+
+      if (! read_full_records_option)
+	{
+	  unsigned long rest = record_size - left;
+
+	  FATAL_ERROR ((0, 0,
+			ngettext ("Unaligned block (%lu byte) in archive",
+				  "Unaligned block (%lu bytes) in archive",
+				  rest),
+			rest));
+	}
+
+      /* User warned us about this.  Fix up.  */
+
+      left -= status;
+      more += status;
+    }
+
+  /* FIXME: for size=0, multi-volume support.  On the first record, warn
+     about the problem.  */
+
+  if (!read_full_records_option && verbose_option > 1
+      && record_start_block == 0 && status != 0)
+    {
+      unsigned long rsize = (record_size - left) / BLOCKSIZE;
+      WARN ((0, 0,
+	     ngettext ("Record size = %lu block",
+		       "Record size = %lu blocks",
+		       rsize),
+	     rsize));
+    }
+
+  record_end = record_start + (record_size - left) / BLOCKSIZE;
+  records_read++;
+}
+
 /* Perform a read to flush the buffer.  */
 void
 flush_read (void)
 {
-  ssize_t status;		/* result from system call */
-  size_t left;			/* bytes left */
-  char *more;			/* pointer to next byte to read */
+  size_t status;		/* result from system call */
 
   if (checkpoint_option && !(++checkpoint % 10))
     WARN ((0, 0, _("Read checkpoint %d"), checkpoint));
@@ -1059,7 +698,7 @@ flush_read (void)
   if (write_archive_to_stdout && record_start_block != 0)
     {
       archive = STDOUT_FILENO;
-      status = write_archive_buffer ();
+      status = sys_write_archive_buffer ();
       archive = STDIN_FILENO;
       if (status != record_size)
 	archive_write_error (status);
@@ -1068,12 +707,7 @@ flush_read (void)
     {
       if (save_name)
 	{
-	  char *cursor = save_name + FILESYSTEM_PREFIX_LEN (save_name);
-
-	  while (ISSLASH (*cursor))
-	    cursor++;
-
-	  assign_string (&real_s_name, cursor);
+	  assign_string (&real_s_name, safer_name_suffix (save_name, false));
 	  real_s_sizeleft = save_sizeleft;
 	  real_s_totsize = save_totsize;
 	}
@@ -1093,9 +727,14 @@ flush_read (void)
       return;
     }
 
+  /* The condition below used to include
+	      || (status > 0 && !read_full_records_option)
+     This is incorrect since even if new_volume() succeeds, the
+     subsequent call to rmtread will overwrite the chunk of data
+     already read in the buffer, so the processing will fail */
+
   if ((status == 0
-       || (status < 0 && errno == ENOSPC)
-       || (status > 0 && !read_full_records_option))
+       || (status == SAFE_READ_ERROR && errno == ENOSPC))
       && multi_volume_option)
     {
       union block *cursor;
@@ -1116,15 +755,12 @@ flush_read (void)
 	  break;
 	}
 
-    vol_error:
-      status = rmtread (archive, record_start->buffer, record_size);
-      if (status < 0)
-	{
-	  archive_read_error ();
-	  goto vol_error;
-	}
+      while ((status = rmtread (archive, record_start->buffer, record_size))
+	     == SAFE_READ_ERROR)
+	archive_read_error ();
+
       if (status != record_size)
-	goto short_read;
+	short_read (status);
 
       cursor = record_start;
 
@@ -1153,7 +789,7 @@ flush_read (void)
 	{
 	  uintmax_t s1, s2;
 	  if (cursor->header.typeflag != GNUTYPE_MULTIVOL
-	      || strcmp (cursor->header.name, real_s_name))
+	      || strncmp (cursor->header.name, real_s_name, NAME_FIELD_SIZE))
 	    {
 	      WARN ((0, 0, _("%s is not continued on this volume"),
 		     quote (real_s_name)));
@@ -1192,46 +828,13 @@ flush_read (void)
       records_read++;
       return;
     }
-  else if (status < 0)
+  else if (status == SAFE_READ_ERROR)
     {
       archive_read_error ();
       goto error_loop;		/* try again */
     }
 
- short_read:
-  more = record_start->buffer + status;
-  left = record_size - status;
-
-  while (left % BLOCKSIZE != 0
-	 || (left && status && read_full_records_option))
-    {
-      if (status)
-	while ((status = rmtread (archive, more, left)) < 0)
-	  archive_read_error ();
-
-      if (status == 0)
-	break;
-
-      if (! read_full_records_option)
-	FATAL_ERROR ((0, 0, _("Unaligned block (%lu bytes) in archive"),
-		      (unsigned long) (record_size - left)));
-
-      /* User warned us about this.  Fix up.  */
-
-      left -= status;
-      more += status;
-    }
-
-  /* FIXME: for size=0, multi-volume support.  On the first record, warn
-     about the problem.  */
-
-  if (!read_full_records_option && verbose_option
-      && record_start_block == 0 && status > 0)
-    WARN ((0, 0, _("Record size = %lu blocks"),
-	   (unsigned long) ((record_size - left) / BLOCKSIZE)));
-
-  record_end = record_start + (record_size - left) / BLOCKSIZE;
-  records_read++;
+  short_read (status);
 }
 
 /*  Flush the current buffer to/from the archive.  */
@@ -1245,17 +848,8 @@ flush_archive (void)
   if (access_mode == ACCESS_READ && time_to_start_writing)
     {
       access_mode = ACCESS_WRITE;
-      time_to_start_writing = 0;
-
-      if (file_to_switch_to >= 0)
-	{
-	  if (rmtclose (archive) != 0)
-	    close_warn (*archive_name_cursor);
-
-	  archive = file_to_switch_to;
-	}
-      else
-	backspace_output ();
+      time_to_start_writing = false;
+      backspace_output ();
     }
 
   switch (access_mode)
@@ -1316,6 +910,42 @@ backspace_output (void)
   }
 }
 
+off_t
+seek_archive (off_t size)
+{
+  off_t start = current_block_ordinal ();
+  off_t offset;
+  off_t nrec, nblk;
+  off_t skipped = (blocking_factor - (current_block - record_start));
+  
+  size -= skipped * BLOCKSIZE;
+  
+  if (size < record_size)
+    return 0;
+  /* FIXME: flush? */
+  
+  /* Compute number of records to skip */
+  nrec = size / record_size;
+  offset = rmtlseek (archive, nrec * record_size, SEEK_CUR);
+  if (offset < 0)
+    return offset;
+
+  if (offset % record_size)
+    FATAL_ERROR ((0, 0, _("rmtlseek not stopped at a record boundary")));
+
+  /* Convert to number of records */
+  offset /= BLOCKSIZE;
+  /* Compute number of skipped blocks */
+  nblk = offset - start;
+
+  /* Update buffering info */
+  records_read += nblk / blocking_factor;
+  record_start_block = offset - blocking_factor;
+  current_block = record_end;
+ 
+  return nblk;
+}
+
 /* Close the archive file.  */
 void
 close_archive (void)
@@ -1323,57 +953,23 @@ close_archive (void)
   if (time_to_start_writing || access_mode == ACCESS_WRITE)
     flush_archive ();
 
-#if !MSDOS
+  sys_drain_input_pipe ();
 
-  /* Manage to fully drain a pipe we might be reading, so to not break it on
-     the producer after the EOF block.  FIXME: one of these days, GNU tar
-     might become clever enough to just stop working, once there is no more
-     work to do, we might have to revise this area in such time.  */
-
-  if (access_mode == ACCESS_READ
-      && ! _isrmt (archive)
-      && (S_ISFIFO (archive_stat.st_mode) || S_ISSOCK (archive_stat.st_mode)))
-    while (rmtread (archive, record_start->buffer, record_size) > 0)
-      continue;
-#endif
-
-  if (verify_option)
+  compute_duration ();
+  if (verify_option) 
     verify_volume ();
 
   if (rmtclose (archive) != 0)
     close_warn (*archive_name_cursor);
 
-#if !MSDOS
+  sys_wait_for_child (child_pid);
 
-  if (child_pid)
-    {
-      int wait_status;
-
-      while (waitpid (child_pid, &wait_status, 0) == -1)
-	if (errno != EINTR)
-	  {
-	    waitpid_error (use_compress_program_option);
-	    break;
-	  }
-
-      if (WIFSIGNALED (wait_status))
-	ERROR ((0, 0, _("Child died with signal %d"),
-		WTERMSIG (wait_status)));
-      else if (WEXITSTATUS (wait_status) != 0)
-	ERROR ((0, 0, _("Child returned status %d"),
-		WEXITSTATUS (wait_status)));
-    }
-#endif /* !MSDOS */
-
-  if (current_file_name)
-    free (current_file_name);
-  if (current_link_name)
-    free (current_link_name);
+  tar_stat_destroy (&current_stat_info);
   if (save_name)
     free (save_name);
   if (real_s_name)
     free (real_s_name);
-  free (multi_volume_option ? record_start - 2 : record_start);
+  free (record_buffer);
 }
 
 /* Called to initialize the global volume number.  */
@@ -1416,9 +1012,10 @@ closeout_volume_number (void)
 }
 
 /* We've hit the end of the old volume.  Close it and open the next one.
-   Return nonzero on success.  */
-static int
-new_volume (enum access_mode access)
+   Return nonzero on success.
+*/
+static bool
+new_volume (enum access_mode mode)
 {
   static FILE *read_file;
   static int looped;
@@ -1428,7 +1025,7 @@ new_volume (enum access_mode access)
     read_file = archive == STDIN_FILENO ? fopen (TTY_NAME, "r") : stdin;
 
   if (now_verifying)
-    return 0;
+    return false;
   if (verify_option)
     verify_volume ();
 
@@ -1489,6 +1086,7 @@ new_volume (enum access_mode access)
 	      {
 	      case '?':
 		{
+		  /* FIXME: Might it be useful to disable the '!' command? */
 		  fprintf (stderr, _("\
  n [name]   Give a new file name for the next (and subsequent) volume(s)\n\
  q          Abort tar\n\
@@ -1516,11 +1114,13 @@ new_volume (enum access_mode access)
 		  char *name = &input_buffer[1];
 		  char *cursor;
 
-		  while (*name == ' ' || *name == '\t')
-		    name++;
-		  cursor = name;
-		  while (*cursor && *cursor != '\n')
-		    cursor++;
+		  for (name = input_buffer + 1;
+		       *name == ' ' || *name == '\t';
+		       name++)
+		    ;
+
+		  for (cursor = name; *cursor && *cursor != '\n'; cursor++)
+		    ;
 		  *cursor = '\0';
 
 		  /* FIXME: the following allocation is never reclaimed.  */
@@ -1529,42 +1129,22 @@ new_volume (enum access_mode access)
 		break;
 
 	      case '!':
-#if MSDOS
-		spawnl (P_WAIT, getenv ("COMSPEC"), "-", 0);
-#else /* not MSDOS */
-		{
-		  pid_t child;
-		  const char *shell = getenv ("SHELL");
-		  if (! shell)
-		    shell = "/bin/sh";
-		  child = xfork ();
-		  if (child == 0)
-		    {
-		      execlp (shell, "-sh", "-i", 0);
-		      exec_fatal (shell);
-		    }
-		  else
-		    {
-		      int wait_status;
-		      while (waitpid (child, &wait_status, 0) == -1)
-			if (errno != EINTR)
-			  {
-			    waitpid_error (shell);
-			    break;
-			  }
-		    }
-		}
-#endif /* not MSDOS */
+		sys_spawn_shell ();
 		break;
 	      }
 	  }
     }
 
-  if (verify_option)
+  if (strcmp (archive_name_cursor[0], "-") == 0)
+    {
+      read_full_records_option = true;
+      archive = STDIN_FILENO;
+    }
+  else if (verify_option)
     archive = rmtopen (*archive_name_cursor, O_RDWR | O_CREAT, MODE_RW,
 		       rsh_command_option);
   else
-    switch (access)
+    switch (mode)
       {
       case ACCESS_READ:
 	archive = rmtopen (*archive_name_cursor, O_RDONLY, MODE_RW,
@@ -1587,14 +1167,13 @@ new_volume (enum access_mode access)
   if (archive < 0)
     {
       open_warn (*archive_name_cursor);
-      if (!verify_option && access == ACCESS_WRITE && backup_option)
+      if (!verify_option && mode == ACCESS_WRITE && backup_option)
 	undo_last_backup ();
       goto tryagain;
     }
 
-#if MSDOS
-  setmode (archive, O_BINARY);
-#endif
+  SET_BINARY_MODE (archive);
 
-  return 1;
+  return true;
 }
+
